from __future__ import annotations

import os
from collections.abc import Callable
from typing import TYPE_CHECKING

from omu.api.server.extension import (
    SERVER_APP_TABLE_TYPE,
    SHUTDOWN_ENDPOINT_TYPE,
    TRUSTED_ORIGINS_REGISTRY_TYPE,
)

from omuserver.session import Session

from .permissions import (
    SERVER_APPS_READ_PERMISSION,
    SERVER_SHUTDOWN_PERMISSION,
    SERVER_TRUSTED_ORIGINS_GET_PERMISSION,
)

if TYPE_CHECKING:
    from loguru import Message

    from omuserver.server import Server


class LogHandler:
    def __init__(
        self,
        callback: Callable[[str], None],
    ) -> None:
        self.callback = callback

    def write(self, message: Message) -> None:
        self.callback(message)


class ServerExtension:
    def __init__(self, server: Server) -> None:
        self._server = server
        server.security.register_permission(
            SERVER_SHUTDOWN_PERMISSION,
            SERVER_APPS_READ_PERMISSION,
            SERVER_TRUSTED_ORIGINS_GET_PERMISSION,
        )
        server.endpoints.bind(SHUTDOWN_ENDPOINT_TYPE, self.handle_shutdown)
        self.apps = self._server.tables.register(SERVER_APP_TABLE_TYPE)
        self.trusted_origins = self._server.registries.register(TRUSTED_ORIGINS_REGISTRY_TYPE)

    async def handle_shutdown(self, session: Session, restart: bool = False) -> bool:
        await self.shutdown(restart)
        return True

    async def shutdown(self, restart: bool = False) -> None:
        try:
            if restart:
                await self._server.restart()
            else:
                await self._server.stop()
        finally:
            os._exit(0)



from __future__ import annotations

import io
from asyncio import Future
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from aiohttp import ClientResponse, ClientSession
from omu.api.http.extension import (
    HTTP_REQUEST_CLOSE,
    HTTP_REQUEST_CREATE,
    HTTP_REQUEST_PERMISSION_ID,
    HTTP_REQUEST_SEND,
    HTTP_RESPONSE_CHUNK,
    HTTP_RESPONSE_CLOSE,
    HTTP_RESPONSE_CREATE,
    HttpChunk,
    HttpRequest,
    HttpResponse,
    RequestHandle,
)
from omu.errors import PermissionDenied
from yarl import URL

from omuserver.api.http.permission import HTTP_REQUEST_PERMISSION
from omuserver.session import Session

if TYPE_CHECKING:
    from omuserver.server import Server


def serialize_response(id: str, response: ClientResponse) -> HttpResponse:
    return {
        "id": id,
        "header": dict(response.headers),
        "history": [serialize_response(id, resp) for resp in response.history],
        "redirected": len(response.history) > 0,
        "status": response.status,
        "statusText": response.reason,
        "url": str(response.url),
    }


@dataclass
class Request:
    session: Session
    buffer: io.BytesIO
    close_future: Future[io.BytesIO] = field(default_factory=Future)


class HttpExtension:
    def __init__(self, server: Server):
        self.server = server
        self.requests: dict[str, Request] = {}
        server.security.register_permission(HTTP_REQUEST_PERMISSION)
        server.network.register_packet(
            HTTP_REQUEST_CREATE,
            HTTP_REQUEST_SEND,
            HTTP_REQUEST_CLOSE,
        )
        server.network.add_packet_handler(HTTP_REQUEST_CREATE, self.handle_request_create)
        server.network.add_packet_handler(HTTP_REQUEST_SEND, self.handle_request_send)
        server.network.add_packet_handler(HTTP_REQUEST_CLOSE, self.handle_request_close)

    async def handle_request_create(self, session: Session, packet: HttpRequest):
        try:
            if not session.permissions.has(HTTP_REQUEST_PERMISSION_ID):
                raise PermissionDenied(f"Missing HTTP request permission: {HTTP_REQUEST_PERMISSION_ID}")
            url = URL(packet["url"])
            request = Request(session=session, buffer=io.BytesIO())
            self.requests[packet["id"]] = request
            await request.close_future
            async with ClientSession() as client:
                async with client.request(
                    packet["method"],
                    url,
                    headers=packet["header"],
                    allow_redirects=packet["redirect"] == "follow",
                    data=request.buffer.getbuffer(),
                ) as response:
                    await session.send(HTTP_RESPONSE_CREATE, serialize_response(packet["id"], response))
                    async for data, _ in response.content.iter_chunks():
                        await session.send(
                            HTTP_RESPONSE_CHUNK,
                            HttpChunk({"id": packet["id"]}, data),
                        )
            await session.send(
                HTTP_RESPONSE_CLOSE,
                {"id": packet["id"]},
            )
        finally:
            del self.requests[packet["id"]]

    async def handle_request_send(self, session: Session, packet: HttpChunk[RequestHandle]):
        handle = self.requests[packet.meta["id"]]
        if handle.session != session:
            raise PermissionDenied("Mismatched session on request handle")
        handle.buffer.write(packet.data)

    async def handle_request_close(self, session: Session, packet: RequestHandle):
        handle = self.requests[packet["id"]]
        if handle.session != session:
            raise PermissionDenied("Mismatched session on request handle")
        handle.close_future.set_result(handle.buffer)



from __future__ import annotations

import json
import time
from asyncio import Future
from typing import TYPE_CHECKING
from venv import logger

from omu.api.dashboard.extension import (
    DASHBOARD_ALLOWED_HOSTS,
    DASHBOARD_APP_INSTALL_ACCEPT_PACKET,
    DASHBOARD_APP_INSTALL_DENY_PACKET,
    DASHBOARD_APP_INSTALL_ENDPOINT,
    DASHBOARD_APP_INSTALL_PACKET,
    DASHBOARD_APP_INSTALL_PERMISSION_ID,
    DASHBOARD_APP_TABLE_TYPE,
    DASHBOARD_APP_UPDATE_ACCEPT_PACKET,
    DASHBOARD_APP_UPDATE_DENY_PACKET,
    DASHBOARD_APP_UPDATE_ENDPOINT,
    DASHBOARD_APP_UPDATE_PACKET,
    DASHBOARD_APP_UPDATE_PERMISSION_ID,
    DASHBOARD_DRAG_DROP_READ_ENDPOINT,
    DASHBOARD_DRAG_DROP_READ_REQUEST_PACKET,
    DASHBOARD_DRAG_DROP_READ_RESPONSE_PACKET,
    DASHBOARD_DRAG_DROP_REQUEST_APPROVAL_PACKET,
    DASHBOARD_DRAG_DROP_REQUEST_ENDPOINT,
    DASHBOARD_DRAG_DROP_REQUEST_PACKET,
    DASHBOARD_DRAG_DROP_STATE_PACKET,
    DASHBOARD_OPEN_APP_ENDPOINT,
    DASHBOARD_OPEN_APP_PACKET,
    DASHBOARD_OPEN_APP_PERMISSION_ID,
    DASHBOARD_PERMISSION_ACCEPT_PACKET,
    DASHBOARD_PERMISSION_DENY_PACKET,
    DASHBOARD_PERMISSION_REQUEST_PACKET,
    DASHBOARD_PLUGIN_ACCEPT_PACKET,
    DASHBOARD_PLUGIN_DENY_PACKET,
    DASHBOARD_PLUGIN_REQUEST_PACKET,
    DASHBOARD_SET_ENDPOINT,
    DASHBOARD_SPEECH_RECOGNITION,
    DASHBOARD_WEBVIEW_EVENT_PACKET,
    DashboardSetResponse,
    DragDropReadRequest,
    DragDropReadResponse,
    DragDropRequest,
    DragDropRequestDashboard,
    DragDropRequestResponse,
    FileDragPacket,
    WebviewEventPacket,
)
from omu.api.dashboard.packets import (
    AppInstallRequestPacket,
    AppInstallResponse,
    AppUpdateRequestPacket,
    AppUpdateResponse,
    PermissionRequestPacket,
    PluginRequestPacket,
)
from omu.app import App, AppType
from omu.errors import PermissionDenied
from omu.identifier import Identifier

from omuserver.session import Session

from .permission import (
    DASHBOARD_APP_INSTALL_PERMISSION,
    DASHBOARD_APP_UPDATE_PERMISSION,
    DASHBOARD_DRAG_DROP_PERMISSION,
    DASHBOARD_OPEN_APP_PERMISSION,
    DASHBOARD_SET_PERMISSION,
    DASHBOARD_SPEECH_RECOGNITION_PERMISSION,
    DASHBOARD_WEBVIEW_PERMISSION,
    DASHOBARD_APP_EDIT_PERMISSION,
    DASHOBARD_APP_READ_PERMISSION,
)

if TYPE_CHECKING:
    from omuserver.server import Server


class DashboardExtension:
    def __init__(self, server: Server) -> None:
        self.server = server
        server.packets.register(
            DASHBOARD_PERMISSION_REQUEST_PACKET,
            DASHBOARD_PERMISSION_ACCEPT_PACKET,
            DASHBOARD_PERMISSION_DENY_PACKET,
            DASHBOARD_PLUGIN_REQUEST_PACKET,
            DASHBOARD_PLUGIN_ACCEPT_PACKET,
            DASHBOARD_PLUGIN_DENY_PACKET,
            DASHBOARD_APP_INSTALL_PACKET,
            DASHBOARD_APP_INSTALL_ACCEPT_PACKET,
            DASHBOARD_APP_INSTALL_DENY_PACKET,
            DASHBOARD_APP_UPDATE_PACKET,
            DASHBOARD_APP_UPDATE_ACCEPT_PACKET,
            DASHBOARD_APP_UPDATE_DENY_PACKET,
            DASHBOARD_DRAG_DROP_STATE_PACKET,
            DASHBOARD_DRAG_DROP_REQUEST_APPROVAL_PACKET,
            DASHBOARD_DRAG_DROP_READ_REQUEST_PACKET,
            DASHBOARD_DRAG_DROP_READ_RESPONSE_PACKET,
            DASHBOARD_WEBVIEW_EVENT_PACKET,
        )
        server.security.register_permission(
            DASHBOARD_SET_PERMISSION,
            DASHBOARD_OPEN_APP_PERMISSION,
            DASHOBARD_APP_READ_PERMISSION,
            DASHOBARD_APP_EDIT_PERMISSION,
            DASHBOARD_APP_INSTALL_PERMISSION,
            DASHBOARD_APP_UPDATE_PERMISSION,
            DASHBOARD_DRAG_DROP_PERMISSION,
            DASHBOARD_WEBVIEW_PERMISSION,
            DASHBOARD_SPEECH_RECOGNITION_PERMISSION,
        )
        server.packets.bind(DASHBOARD_PERMISSION_ACCEPT_PACKET, self.handle_permission_accept)
        server.packets.bind(DASHBOARD_PERMISSION_DENY_PACKET, self.handle_permission_deny)
        server.packets.bind(DASHBOARD_PLUGIN_ACCEPT_PACKET, self.handle_plugin_accept)
        server.packets.bind(DASHBOARD_PLUGIN_DENY_PACKET, self.handle_plugin_deny)
        server.packets.bind(DASHBOARD_APP_INSTALL_ACCEPT_PACKET, self.handle_app_install_accept)
        server.packets.bind(DASHBOARD_APP_INSTALL_DENY_PACKET, self.handle_app_install_deny)
        server.packets.bind(DASHBOARD_APP_UPDATE_ACCEPT_PACKET, self.handle_app_update_accept)
        server.packets.bind(DASHBOARD_APP_UPDATE_DENY_PACKET, self.handle_app_update_deny)
        server.packets.bind(DASHBOARD_DRAG_DROP_STATE_PACKET, self.handle_drag_drop_state)
        server.packets.bind(DASHBOARD_DRAG_DROP_REQUEST_APPROVAL_PACKET, self.handle_drag_drop_request_approval)
        server.packets.bind(DASHBOARD_DRAG_DROP_READ_RESPONSE_PACKET, self.handle_drag_drop_read_response)
        server.packets.bind(DASHBOARD_WEBVIEW_EVENT_PACKET, self.handle_webview_event)
        server.endpoints.bind(DASHBOARD_SET_ENDPOINT, self.handle_dashboard_set)
        server.endpoints.bind(DASHBOARD_OPEN_APP_ENDPOINT, self.handle_dashboard_open_app)
        server.endpoints.bind(DASHBOARD_APP_INSTALL_ENDPOINT, self.handle_dashboard_app_install)
        server.endpoints.bind(DASHBOARD_APP_UPDATE_ENDPOINT, self.handle_dashboard_app_update)
        server.endpoints.bind(DASHBOARD_DRAG_DROP_REQUEST_ENDPOINT, self.handle_drag_drop_request)
        server.endpoints.bind(DASHBOARD_DRAG_DROP_READ_ENDPOINT, self.handle_drag_drop_read)
        self.apps = server.tables.register(DASHBOARD_APP_TABLE_TYPE)
        self.allowed_hosts = server.tables.register(DASHBOARD_ALLOWED_HOSTS)
        self.speech_recognition = server.registries.register(DASHBOARD_SPEECH_RECOGNITION)
        self.dashboard_session: Session | None = None
        self.dashboard_wait_future: Future[Session] | None = None
        self.permission_requests: dict[str, Future[bool]] = {}
        self.plugin_requests: dict[str, Future[bool]] = {}
        self.app_install_requests: dict[str, Future[bool]] = {}
        self.app_update_requests: dict[str, Future[bool]] = {}
        self.drag_drop_requests: dict[str, Future[DragDropRequestResponse]] = {}
        self.drag_drop_sessions: dict[str, Session] = {}
        self.drag_drop_states: dict[str, Session] = {}
        self.drag_drop_read_requests: dict[str, Future[DragDropReadResponse]] = {}
        self.request_id = 0

    async def wait_dashboard_ready(self) -> Session:
        if self.dashboard_session:
            return self.dashboard_session
        if self.dashboard_wait_future:
            return await self.dashboard_wait_future
        self.dashboard_wait_future = Future()
        return await self.dashboard_wait_future

    async def handle_session_connected(self, session: Session) -> None:
        if session.kind != AppType.APP:
            return
        exist_app = await self.apps.get(session.app.id.key())
        if exist_app is None:
            return
        old_metadata = json.dumps(exist_app.metadata)
        new_metadata = json.dumps(session.app.metadata)
        if old_metadata != new_metadata:
            await self.update_app(session.app)

    async def handle_dashboard_open_app(self, session: Session, app: App) -> None:
        if self.dashboard_session is None:
            raise ValueError("Dashboard session not set")
        if not session.permissions.has(DASHBOARD_OPEN_APP_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to open apps")
        await self.dashboard_session.send(DASHBOARD_OPEN_APP_PACKET, app)

    async def handle_dashboard_set(self, session: Session, identifier: Identifier) -> DashboardSetResponse:
        if session.kind != AppType.DASHBOARD:
            raise PermissionDenied("Session is not a dashboard")
        self.dashboard_session = session
        session.event.disconnected += self._on_dashboard_disconnected
        if self.dashboard_wait_future:
            self.dashboard_wait_future.set_result(session)
            self.dashboard_wait_future = None
        return {"success": True}

    async def _on_dashboard_disconnected(self, session: Session) -> None:
        self.dashboard_session = None

    def ensure_dashboard_session(self, session: Session) -> bool:
        if session == self.dashboard_session:
            return True
        msg = f"Session {session} is not the dashboard session"
        raise PermissionDenied(msg)

    async def handle_permission_accept(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.permission_requests:
            raise ValueError(f"Permission request with id {request_id} does not exist")
        future = self.permission_requests.pop(request_id)
        future.set_result(True)

    async def handle_permission_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.permission_requests:
            raise ValueError(f"Permission request with id {request_id} does not exist")
        future = self.permission_requests.pop(request_id)
        future.set_result(False)

    async def request_permissions(self, request: PermissionRequestPacket) -> bool:
        if request.request_id in self.permission_requests:
            raise ValueError(f"Permission request with id {request.request_id} already exists")
        future = Future[bool]()
        self.permission_requests[request.request_id] = future
        dashboard = await self.wait_dashboard_ready()
        await dashboard.send(
            DASHBOARD_PERMISSION_REQUEST_PACKET,
            request,
        )
        return await future

    async def handle_plugin_accept(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.plugin_requests:
            raise ValueError(f"Plugin request with id {request_id} does not exist")
        future = self.plugin_requests.pop(request_id)
        future.set_result(True)

    async def handle_plugin_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.plugin_requests:
            raise ValueError(f"Plugin request with id {request_id} does not exist")
        future = self.plugin_requests.pop(request_id)
        future.set_result(False)

    async def request_plugins(self, request: PluginRequestPacket) -> bool:
        if request.request_id in self.plugin_requests:
            raise ValueError(f"Plugin request with id {request.request_id} already exists")
        future = Future[bool]()
        self.plugin_requests[request.request_id] = future
        dashboard = await self.wait_dashboard_ready()
        await dashboard.send(
            DASHBOARD_PLUGIN_REQUEST_PACKET,
            request,
        )
        return await future

    async def handle_app_install_accept(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_install_requests:
            raise ValueError(f"App install request with id {request_id} does not exist")
        future = self.app_install_requests.pop(request_id)
        future.set_result(True)

    async def handle_app_install_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_install_requests:
            raise ValueError(f"App install request with id {request_id} does not exist")
        future = self.app_install_requests.pop(request_id)
        future.set_result(False)

    async def handle_dashboard_app_install(self, session: Session, app: App) -> AppInstallResponse:
        if not session.permissions.has(DASHBOARD_APP_INSTALL_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to install apps")
        request_id = self.gen_next_request_id()
        future = Future[bool]()
        self.app_install_requests[request_id] = future
        dashboard = await self.wait_dashboard_ready()
        request = AppInstallRequestPacket(request_id=request_id, app=app)
        await dashboard.send(
            DASHBOARD_APP_INSTALL_PACKET,
            request,
        )
        accepted = await future
        if accepted:
            await self.apps.add(app)
        return AppInstallResponse(accepted=accepted)

    async def handle_app_update_accept(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_update_requests:
            raise ValueError(f"App update request with id {request_id} does not exist")
        future = self.app_update_requests.pop(request_id)
        future.set_result(True)

    async def handle_app_update_deny(self, session: Session, request_id: str) -> None:
        self.ensure_dashboard_session(session)
        if request_id not in self.app_update_requests:
            raise ValueError(f"App update request with id {request_id} does not exist")
        future = self.app_update_requests.pop(request_id)
        future.set_result(False)

    async def update_app(self, app: App) -> bool:
        old_app = await self.apps.get(app.id.key())
        if old_app is None:
            raise ValueError(f"App with id {app.id} does not exist")
        request_id = self.gen_next_request_id()
        future = Future[bool]()
        self.app_update_requests[request_id] = future
        dashboard = await self.wait_dashboard_ready()
        await dashboard.send(
            DASHBOARD_APP_UPDATE_PACKET,
            AppUpdateRequestPacket(
                request_id=request_id,
                old_app=old_app,
                new_app=app,
            ),
        )
        accepted = await future
        if accepted:
            await self.apps.add(app)
        return accepted

    async def handle_dashboard_app_update(self, session: Session, app: App) -> AppUpdateResponse:
        if not session.permissions.has(DASHBOARD_APP_UPDATE_PERMISSION_ID):
            raise PermissionDenied("Session does not have permission to update apps")
        accepted = await self.update_app(app)
        return AppUpdateResponse(accepted=accepted)

    async def handle_drag_drop_state(self, session: Session, packet: FileDragPacket):
        self.ensure_dashboard_session(session)
        app = packet["app"]
        target = self.drag_drop_sessions.get(app["id"], None)
        if target is None:
            logger.warning(
                f"Dashboard {self.dashboard_session} tried to send drag drop state to unknown app {app['id']}"
            )
            return
        self.drag_drop_states[packet["drag_id"]] = target
        await target.send(DASHBOARD_DRAG_DROP_STATE_PACKET, packet)

    async def handle_drag_drop_request_approval(self, session: Session, response: DragDropRequestResponse):
        self.ensure_dashboard_session(session)
        request = self.drag_drop_requests.pop(response["request_id"], None)
        if not request:
            msg = (
                f"Dashboard {self.dashboard_session} tried to {"approve" if response['ok'] else "deny"} "
                f"to unknown request {response['request_id']} with"
            )
            logger.warning(msg)
            return
        request.set_result(response)

    async def handle_drag_drop_request(self, session: Session, request: DragDropRequest) -> DragDropRequestResponse:
        dashboard = await self.wait_dashboard_ready()
        request_id = self.gen_next_request_id()
        request = DragDropRequestDashboard(
            request_id=request_id,
            app=session.app.to_json(),
        )
        await dashboard.send(DASHBOARD_DRAG_DROP_REQUEST_PACKET, request)
        future = Future[DragDropRequestResponse]()
        self.drag_drop_requests[request_id] = future
        response = await future
        if response["ok"]:
            self.drag_drop_sessions[session.app.key()] = session
        return response

    async def handle_drag_drop_read(self, session: Session, request: DragDropReadRequest) -> DragDropReadResponse:
        dashboard = await self.wait_dashboard_ready()
        target = self.drag_drop_states[request["drag_id"]]
        if target != session:
            msg = f"Session {session} tried to read invalid drag drop file"
            logger.warning(msg)
            raise PermissionDenied(msg)
        request_id = self.gen_next_request_id()
        await dashboard.send(
            DASHBOARD_DRAG_DROP_READ_REQUEST_PACKET,
            {"drag_id": request["drag_id"], "request_id": request_id},
        )
        future = Future[DragDropReadResponse]()
        self.drag_drop_read_requests[request_id] = future
        return await future

    async def handle_drag_drop_read_response(self, session: Session, response: DragDropReadResponse):
        drag_id = response.meta["request_id"]
        request = self.drag_drop_read_requests[drag_id]
        request.set_result(response)

    async def handle_webview_event(self, session: Session, packet: WebviewEventPacket):
        self.ensure_dashboard_session(session)
        target = self.server.sessions.find(packet.target)
        if target is None:
            return
        await target.send(DASHBOARD_WEBVIEW_EVENT_PACKET, packet)

    def gen_next_request_id(self) -> str:
        self.request_id += 1
        return f"{self.request_id}-{time.time_ns()}"



from __future__ import annotations

import asyncio
import json
import os
import subprocess
import sys
from urllib.parse import urlencode

import aiohttp
from aiohttp import web
from loguru import logger
from omu import Identifier
from omu.event_emitter import EventEmitter
from omu.helper import asyncio_error_logger
from omu.network.packet.packet_types import DisconnectType
from yarl import URL

from omuserver.api.asset import AssetExtension
from omuserver.api.dashboard import DashboardExtension
from omuserver.api.endpoint import EndpointExtension
from omuserver.api.http import HttpExtension
from omuserver.api.i18n import I18nExtension
from omuserver.api.logger import LoggerExtension
from omuserver.api.permission import PermissionExtension
from omuserver.api.plugin import PluginExtension
from omuserver.api.registry import RegistryExtension
from omuserver.api.server import ServerExtension
from omuserver.api.session import SessionExtension
from omuserver.api.signal import SignalExtension
from omuserver.api.table import TableExtension
from omuserver.config import Config
from omuserver.helper import safe_path_join
from omuserver.network import Network
from omuserver.network.packet_dispatcher import ServerPacketDispatcher
from omuserver.security import PermissionManager
from omuserver.version import VERSION

USER_AGENT_HEADERS = {"User-Agent": json.dumps(["omu", {"name": "omuserver", "version": VERSION}])}
RESTART_EXIT_CODE = 100
FRAME_TYPE_KEY = "omuapps-frame"


class Server:
    def __init__(
        self,
        config: Config,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> None:
        self.config = config
        self._loop = self._set_loop(loop or asyncio.new_event_loop())
        self.address = config.address
        self.event = ServerEvents()
        self.directories = config.directories
        self.directories.mkdir()
        self.packets = ServerPacketDispatcher()
        self.network = Network(self, self.packets)
        self.network.add_http_route("/", self._handle_index)
        self.network.add_http_route("/version", self._handle_version)
        self.network.add_http_route("/frame", self._handle_frame)
        self.network.add_http_route("/proxy", self._handle_proxy)
        self.network.add_http_route("/asset", self._handle_assets)
        self.security = PermissionManager.load(self)
        self.running = False
        self.endpoints = EndpointExtension(self)
        self.permissions = PermissionExtension(self)
        self.tables = TableExtension(self)
        self.sessions = SessionExtension(self)
        self.http = HttpExtension(self)
        self.registries = RegistryExtension(self)
        self.dashboard = DashboardExtension(self)
        self.server = ServerExtension(self)
        self.signals = SignalExtension(self)
        self.plugins = PluginExtension(self)
        self.assets = AssetExtension(self)
        self.i18n = I18nExtension(self)
        self.logger = LoggerExtension(self)
        self.client = aiohttp.ClientSession(
            loop=self.loop,
            headers=USER_AGENT_HEADERS,
            timeout=aiohttp.ClientTimeout(total=10),
        )

    def _set_loop(self, loop: asyncio.AbstractEventLoop) -> asyncio.AbstractEventLoop:
        loop = asyncio.new_event_loop()
        loop.set_exception_handler(asyncio_error_logger)
        return loop

    async def _handle_index(self, request: web.Request) -> web.StreamResponse:
        return web.FileResponse(self.directories.index)

    async def _handle_version(self, request: web.Request) -> web.Response:
        return web.json_response({"version": VERSION})

    async def _handle_frame(self, request: web.Request) -> web.StreamResponse:
        url = request.query.get("url")
        if not url:
            return web.Response(status=400)
        url = URL(url).human_repr()
        content = self.directories.frame.read_text(encoding="utf-8")
        frame_token = self.security.generate_frame_token(url)
        config = {
            "frame_token": frame_token,
            "url": url,
            "ws_url": URL.build(
                scheme="ws",
                host=self.address.host,
                port=self.address.port,
                path="/ws",
                query_string=urlencode({"frame_token": frame_token, "url": url}),
            ).human_repr(),
            "type_key": FRAME_TYPE_KEY,
        }
        content = content.replace("%CONFIG%", json.dumps(config))
        return web.Response(text=content, content_type="text/html")

    async def _handle_proxy(self, request: web.Request) -> web.StreamResponse:
        url = request.query.get("url")
        no_cache = bool(request.query.get("no_cache"))
        if not url:
            return web.Response(status=400)
        try:
            async with self.client.get(
                url,
            ) as resp:
                headers = {
                    "Cache-Control": "no-cache" if no_cache else "max-age=3600",
                    "Content-Type": resp.content_type,
                    "Access-Control-Allow-Origin": "*",
                }
                response = web.StreamResponse(status=resp.status, headers=headers)
                await response.prepare(request)
                async for chunk in resp.content.iter_any():
                    await response.write(chunk)
                return response
        except TimeoutError:
            return web.Response(status=504)
        except aiohttp.ClientConnectionResetError:
            return web.Response(status=502)
        except aiohttp.ClientResponseError as e:
            return web.Response(status=e.status, text=e.message)
        except Exception:
            logger.error("Failed to proxy request")
            return web.Response(status=500)

    async def _handle_assets(self, request: web.Request) -> web.StreamResponse:
        id = request.query.get("id")
        if not id:
            return web.Response(status=400)
        identifier = Identifier.from_key(id)
        path = identifier.get_sanitized_path()
        try:
            path = safe_path_join(self.directories.assets, path)

            if not path.exists():
                return web.Response(status=404)
            return web.FileResponse(path)
        except Exception as e:
            logger.error(e)
            return web.Response(status=500)

    def run(self) -> None:
        async def _run():
            await self.start()

        if self._loop is None:
            asyncio.run(_run())
        else:
            self._loop.create_task(_run())
            self._loop.run_forever()

    async def start(self) -> None:
        self.running = True
        try:
            await self.network.start()
            logger.info(f"Listening on {self.address.host}:{self.address.port}")
            await self.event.start()
        except Exception as e:
            logger.opt(exception=e).error("Failed to start server")
            await self.stop()
            self.loop.stop()
            raise e

    async def stop(self) -> None:
        logger.info("Stopping server")
        self.running = False
        await self.event.stop()
        await self.network.stop()

    async def restart(self) -> None:
        for session in list(self.sessions.iter()):
            if session.closed:
                continue
            await session.disconnect(DisconnectType.SERVER_RESTART, "Server is restarting")
        await self.stop()
        child = subprocess.Popen(
            args=[sys.executable, "-m", "omuserver", *sys.argv[1:]],
            cwd=os.getcwd(),
        )
        logger.info(f"Restarting server with PID {child.pid}")
        os._exit(RESTART_EXIT_CODE)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop


class ServerEvents:
    def __init__(self) -> None:
        self.start = EventEmitter[[]]()
        self.stop = EventEmitter[[]]()
